#include <Piece.h>
#include <PieceFactory.h>
#include <PieceComparator.h>
#include <PieceType.h>
#include <Types.h>

#include <vector>
#include <queue>
#include <string>

using namespace std;

/*
* The virtual destructor.
*/
PieceFactory::~PieceFactory() {
	release_memory();
}

/*
* Release the memory.
*/
void PieceFactory::release_memory() {    
    x_pieces.clear();
    y_pieces.clear();
    z_pieces.clear();
    used_pieces.clear();
    all_pieces.clear();
}

/*
* This method creates a new piece and stores it in the proper place.
*/
Piece * PieceFactory::new_piece(
    void * var_info, 
    PieceType type, 
    unsigned size, 
    unsigned priority,
    unsigned preference,
    unsigned span){
    // A pointer to the needed map.
    map<
        unsigned,
        map<
            unsigned,
            priority_queue<
                Piece *,
                vector<Piece *>,
                PieceComparator> *
            > 
        > * queue_map;
    // A pointer to the needed queue.
    priority_queue<
        Piece *,
        vector<Piece *>,
        PieceComparator
        > * queue;
        
    switch(type){
        case X:
            queue_map = &x_pieces;
        break;
        case Y:
            queue_map = &y_pieces;
        break;
        case Z:
            queue_map = &z_pieces;
    }
            
    // We create the piece
    Piece * piece = new Piece(var_info,type,size,priority);
    //If the piece cannot be created then an exception is thrown
    if(!(*queue_map)[preference][span]){
        (*queue_map)[preference][span] = new priority_queue<
            Piece *,
            vector<Piece *>,
            PieceComparator>();
    }
    (*queue_map)[preference][span]->push(piece);
    all_pieces.push_back(piece);
    return piece;
}

/*
* Returns a piece of type, preference and span given.
* Store the piece inside used_pieces and deletes 
* the piece from the available pieces.
*/
Piece * PieceFactory::get_piece(
    PieceType type, 
    unsigned preference,
    unsigned span){
    switch(type){
      case X:
        if(x_pieces[preference][span] && 
        ! (x_pieces[preference][span]->empty())){
            Piece * piece = x_pieces[preference][span]->top();
            x_pieces[preference][span]->pop();
            used_pieces.push_back(piece);
            return piece;
        }   
        break;
      case Y:
        if(y_pieces[preference][span] && 
        ! (y_pieces[preference][span]->empty())){
            Piece * piece = y_pieces[preference][span]->top();
            y_pieces[preference][span]->pop();
            used_pieces.push_back(piece);
            return piece;
        }
      break;
      case Z:
        if(z_pieces[preference][span] && 
        ! (z_pieces[preference][span]->empty())){
            Piece * piece = z_pieces[preference][span]->top();
            z_pieces[preference][span]->pop();
            used_pieces.push_back(piece);
            return piece;
        }
    }
    return 0;
}

/*
* This method returns a piece on the specified preferences.
* preferences_begin is an iterator to the biginning of the preferences 
* vector.
* preferences_end is an iterator to the end of the preferences 
* vector.
* Those iterators should be the ones retrived for the call
* to Area::get_preferences_begin and Area::get_preferences_end.
*/
Piece * PieceFactory::get_piece(
    PieceType type, 
    vector<unsigned>::iterator preferences_begin,
    vector<unsigned>::iterator preferences_end,
    unsigned span){
    for(;preferences_begin != preferences_end;preferences_begin++){
        if(get_available_number(type,*preferences_begin,span)>=1){
            return get_piece(type,*preferences_begin,span);
        }
    }
    return 0;
}

/*
* This method returns how many Pieces are available to use.
* with the specified type 
* The preferences_begin argument is an iterator to the begin of a vector 
* of preferences. It should be the one in Area class.
* Since the Area subclasses, the iterators can be obtained with 
* the get_preferences_begin method.
* The preferences_end argument is an iterator to the begin of a vector 
* of preferences. It should be the one in Area class
* Since the Area subclasses, the iterators can be obtained with 
* the get_preferences_end method.
*/
unsigned PieceFactory::get_available_number(
    PieceType type, 
    vector<unsigned>::iterator preferences_begin,
    vector<unsigned>::iterator preferences_end,
    unsigned span){
    unsigned available = 0;
    for(;preferences_begin != preferences_end;preferences_begin++){
        available += get_available_number(type,*preferences_begin,span);
    }
    return available;
}

/*
* Returns the available number of pieces with preference and span given.
*/
unsigned PieceFactory::get_available_number(
    PieceType type,     
    unsigned preference,
    unsigned span){
    unsigned available = 0 ;
    switch(type){
      case X:
        if(x_pieces[preference][span])
            available = x_pieces[preference][span]->size();
        break;
      case Y:
        if(y_pieces[preference][span])
            available = y_pieces[preference][span]->size();
        break;
      case Z:
        if(z_pieces[preference][span])
            available = z_pieces[preference][span]->size();
        break;
    }
    return available;
}

/*
* This method returns the total number of available pieces.
*/
unsigned PieceFactory::get_available_number(){
    return all_pieces.size()-used_pieces.size();
}

/*
* Returns the number of x spots that can be filled with the non
* used pieces in this PieceFactory.
* Also with Y pieces.
*/
unsigned PieceFactory::get_x_to_fill_number(){
    unsigned result = 0;
    //The needed iterators
    map<
        unsigned,
        map<
            unsigned,
            priority_queue<
                Piece *,
                vector<Piece *>,
                PieceComparator> *
            > 
        >::iterator pieces_it;
    
    map<
        unsigned,
        priority_queue<
            Piece *,
            vector<Piece * >, 
            PieceComparator > *
        >::iterator span_it;
    
    //Go to x pieces
    for ( pieces_it=x_pieces.begin();pieces_it != x_pieces.end();pieces_it++ ){
        for(span_it=(*pieces_it).second.begin();
            span_it!=(*pieces_it).second.end();
            span_it++){
            result += ( (*span_it).first ) * ( (*span_it).second -> size() );
        }
    }
    
    //Go to y pieces
    for ( pieces_it=y_pieces.begin();pieces_it != y_pieces.end();pieces_it++ ){
        for(span_it=(*pieces_it).second.begin();
            span_it!=(*pieces_it).second.end();
            span_it++){
            result += ( (*span_it).first ) * ( (*span_it).second -> size() );
        }
    }
    
    return result;

}

/*
* Returns the number of z spots that can be filled with the non
* used pieces in this PieceFactory.
* Also with Y pieces.    
*/
unsigned PieceFactory::get_z_to_fill_number(){
    unsigned result = 0;
    //The needed iterators
    map<
        unsigned,
        map<
            unsigned,
            priority_queue<
                Piece *,
                vector<Piece *>,
                PieceComparator> *
            > 
        >::iterator pieces_it;
    
    map<
        unsigned,
        priority_queue<Piece *,
        vector<Piece * >, 
        PieceComparator > *
        >::iterator span_it;
    
    //Go to z pieces
    for ( pieces_it=z_pieces.begin();pieces_it != z_pieces.end();pieces_it++ ){
        for(span_it=(*pieces_it).second.begin();
            span_it!=(*pieces_it).second.end();
            span_it++){
            result += ( (*span_it).first ) * ( (*span_it).second -> size() );
        }
    }
    
    //Go to y pieces
    for ( pieces_it=y_pieces.begin();pieces_it != y_pieces.end();pieces_it++ ){
        for(span_it=(*pieces_it).second.begin();
            span_it!=(*pieces_it).second.end();
            span_it++){
            result += ( (*span_it).first ) * ( (*span_it).second -> size() );
        }
    }
    
    return result;

}

string PieceFactory::all_to_string(DATA_STRING var_info_to_string){
    string rep = "" ;
    int i;
    for ( i = 0 ; i < all_pieces.size() ; i++){
        rep += all_pieces.at(i)->to_string(var_info_to_string) + "\n";
    }
    return rep;
}


string PieceFactory::used_to_string(DATA_STRING var_info_to_string){
    string rep = "" ;
    int i;
    //Revisar todas las piezas
    for ( i = 0 ; i < used_pieces.size() ; i++){
        rep += used_pieces.at(i)->to_string(var_info_to_string) + "\n";
    }
    return rep;
}
