#ifndef _PIECE_FACTORY_H
#define _PIECE_FACTORY_H

#include <Piece.h>
#include <PieceComparator.h>

#include <map>
#include <queue>
#include <string>

using namespace std;

class PieceFactory {

  public:
  
    /*
    * The virtual destructor.
    */
    virtual ~PieceFactory();

    /*
    * Release memory.
    */
    virtual void release_memory();

    /*
    * Creates a new piece.
    */
    Piece * new_piece(
        void * var_info, 
        PieceType type,
        unsigned size, 
        unsigned priority,
        unsigned preference,
        unsigned span);

    /*
    * Returns a pointer to the piece that has the given variable information.
    */
    Piece * get_piece(void * var_info);
    
    /*
    * This method returns a pointer to a piece of a given type and
    * preference.
    * If it is not posible, the method returns a null pointer.
    */
    Piece * get_piece(
        PieceType type, 
        unsigned preference,
        unsigned span);
        
    /*
    * This method returns a piece on the specified preferences.
    * preferences_begin is an iterator to the biginning of the preferences 
    * vector.
    * preferences_end is an iterator to the end of the preferences 
    * vector.
    * Those iterators should be the ones retrived for the call
    * to Area::get_preferences_begin and Area::get_preferences_end.
    */
    Piece * get_piece(
        PieceType type, 
        vector<unsigned>::iterator preferences_begin,
        vector<unsigned>::iterator preferences_end,
        unsigned span);
    
    /*
    * This method returns how many Pieces are available to use.
    * with the specified type, preference and span.
    */
    unsigned get_available_number(
        PieceType type, 
        unsigned preference,
        unsigned span);
        
    /*
    * This method returns how many Pieces are available to use.
    * with the specified type 
    * The preferences_begin argument is an iterator to the begin of a vector 
    * of preferences. It should be the one in Area class.
    * Since the Area subclasses, the iterators can be obtained with 
    * the get_preferences_begin method.
    * The preferences_end argument is an iterator to the begin of a vector 
    * of preferences. It should be the one in Area class
    * Since the Area subclasses, the iterators can be obtained with 
    * the get_preferences_end method.
    */
    unsigned get_available_number(
        PieceType type, 
        vector<unsigned>::iterator preferences_begin,
        vector<unsigned>::iterator preferences_end,
        unsigned span);
    
    /*
    * This method returns the total number of available pieces.
    */
    unsigned get_available_number();
    
    /*
    * Returns the number of x spots that can be filled with the non
    * used pieces in this PieceFactory.
    * Also with Y pieces.
    */
    unsigned get_x_to_fill_number();
    
    /*
    * Returns the number of z spots that can be filled with the non
    * used pieces in this PieceFactory.
    * Also with Y pieces.    
    */
    unsigned get_z_to_fill_number();
    
    /*
    * This method returns a string representation of all the pieces.
    */
    virtual string all_to_string(DATA_STRING var_info_to_string);
    
    /*
    * This method returns a string representation of the used pieces.
    */
    virtual string used_to_string(DATA_STRING var_info_to_string);

  protected:
    /*
    * This implementation gives support to future puzzle solver algorithms.
    * The pieces are stored in priority queues in order to be able to implement
    * some algorithm that produces pieces with more priority than others, 
    * it should make the register allocation more intelligent.
    * The first map level defines the preference
    * The second map level defines the span of the piece
    */
    map<
        unsigned,
        map<
            unsigned,
            priority_queue<
                Piece *,
                vector<Piece *>,
                PieceComparator> *
            > 
        > x_pieces;
    
    /*
    * To store the pointers of each y piece created.
    */
    map<
        unsigned,
        map<
            unsigned,
            priority_queue<
                Piece *,
                vector<Piece *>,
                PieceComparator> *
            > 
        > y_pieces;
    
    /*
    * To store the pointers of each z piece created.
    */
    map<
        unsigned,
        map<
            unsigned,
            priority_queue<
                Piece *,
                vector<Piece *>,
                PieceComparator> *
            > 
        > z_pieces;
    
    /*
    * This vector will store the used pieces.
    */
    vector<Piece *> used_pieces;

    /*
    * This vector will contain all the pieces created.
    * It is useful only because there are no iterators in the priority_queue
    * class, and to be able to know more information about all the pieces, we 
    * need to store them somewhere else.
    */
    vector<Piece *> all_pieces;

};

#endif
